#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "dict.h"

void dict_AddChineseFast(English2Chinese * E2C,char * chinese)
{
  E2C->chinese = (char**)realloc(E2C->chinese,sizeof(char*)*(E2C->size+1));
  E2C->chinese[E2C->size] = strdup(chinese);
  E2C->size++;
}

void dict_AddChinese(English2Chinese * E2C,char *chinese)
{
  int n;
  
  for (n=0;n<E2C->size;n++)
    if (strcmp(chinese,E2C->chinese[n])==0) return;

  dict_AddChineseFast(E2C,chinese);

}

int dict_BinarySearchEnglish(E2C_Dictionary * DICT,char *english,int left,int right)
{
  int mid = (left+right)/2;
  int Lanswer,Manswer,Ranswer;

  if (left>right) return -1;
  mid = (left+right)/2;
  Lanswer = strcasecmp(english,DICT->DICT[left]->english);
  if (Lanswer==0) return left;
  Manswer = strcasecmp(english,DICT->DICT[mid]->english);
  if (Manswer==0) return mid;
  Ranswer = strcasecmp(english,DICT->DICT[right]->english);
  if (Ranswer==0) return right;
  if (Lanswer*Manswer<0) 
    return dict_BinarySearchEnglish(DICT,english,left+1,mid-1);
  if (Manswer*Ranswer<0)
    return dict_BinarySearchEnglish(DICT,english,mid+1,right-1);
  return -1;
}

int dict_SearchEnglish(E2C_Dictionary * DICT,char * english)
{
  return dict_BinarySearchEnglish(DICT,english,0,DICT->size-1);
}

int dict_FindEnglishGap(E2C_Dictionary * DICT, char * english,int left,int right)
{
  int mid;
  int Lanswer;
  int Manswer;
  int Ranswer;
  
  if (left>right) return -1;
  Lanswer = strcasecmp(english,DICT->DICT[left]->english);
  if (Lanswer==0) return left;
  Ranswer = strcasecmp(english,DICT->DICT[right]->english);
  if (Ranswer==0) return right;
  if (Lanswer*Ranswer>0) return -1;
  if ((right-left)==1) return right;
  mid = (left+right)/2;
  Manswer = strcasecmp(english,DICT->DICT[mid]->english);
  if (Lanswer*Manswer<0) {
    if ((mid-left)==1) return mid;
    return dict_FindEnglishGap(DICT,english,left,mid);
  } else {
    if ((right-mid)==1) return right;
    return dict_FindEnglishGap(DICT,english,mid,right);
  }
  return -1;
}

int dict_InsertEnglish(E2C_Dictionary * DICT,char * english)
{
  int n;
  int id;


  id = dict_FindEnglishGap(DICT,english,0,DICT->size-1);
  if (id<0) id = DICT->size;
  DICT->DICT = (English2Chinese **)realloc(
                  DICT->DICT,
                  sizeof(English2Chinese*)*(DICT->size+1)
                );
  /* move over */
  n = DICT->size;
  while (n>id) {
    DICT->DICT[n] = DICT->DICT[n-1];
    n--;
  }
  DICT->DICT[id] = (English2Chinese *)malloc(sizeof(English2Chinese));
  memset(DICT->DICT[id],0,sizeof(English2Chinese));
  DICT->DICT[id]->english = strdup(english);
  DICT->size++;
  
  return id;
}

char * dict_FindEnglish(char *buff,char **next)
{
  char * p;
  /* find first quote */
  while (*buff&&*buff!='"') buff++;
  if (*buff==0) return NULL;
  buff++;
  if (*buff==0) return NULL;
  p = buff;
  /* find second quote */
  while (*buff&&*buff!='"') buff++;
  if (*buff==0) return NULL;
  *buff = 0;
  buff++;
  *next = buff;
  
  return p;
}

char * dict_FindChinese(char *buff,char **next)
{
  char * p;
  
  while (*buff&&(*buff==' '||*buff==9)) buff++;
  if (*buff==0) return NULL;
  p = buff;
  while (*buff&&*buff!=' '&&*buff!=9&&*buff!=10&&*buff!=13) buff++;
  *buff = 0; buff++;
  *next = buff;
  return p;
}

int dict_LoadDictionary(FILE * fp,E2C_Dictionary * DICT)
{
  int n,w;
  char buff[4096];
  English2Chinese * word;
  char * first;
  char * next;

  while (fgets(buff,4096,fp)!=NULL) {
    first = dict_FindEnglish(buff,&next);
    if (first==NULL) continue;
    n = dict_SearchEnglish(DICT,first);
    if (n<0) n = dict_InsertEnglish(DICT,first);
    word = DICT->DICT[n];
    do {
      first = dict_FindChinese(next,&next);
      if (first!=NULL) dict_AddChinese(word,first);
    } while (first!=NULL);
  }
  return DICT->size;
}

static int ref_add(E2C_Dictionary * REF,English2Chinese *Word)
{
  int n;

  for (n=0;n<REF->size;n++) if (REF->DICT[n]==Word) return REF->size;

  REF->DICT = (English2Chinese **)realloc(
                 REF->DICT,
                 sizeof(English2Chinese*)*(REF->size+1)
              );
  REF->DICT[REF->size] = Word;
  REF->size++;
  return REF->size;
}

static char * ref_findword(char * p,char **next)
{
  char buff[4096];
  int n;
  int ok;

  ok = 0;
  do {
    if (*p) {
      if ((*p>='a'&&*p<='z')||(*p>='A'&&*p<='Z')) ok=1; else p++;
    } else ok=1;
  } while (!ok);

  if (*p==0) return NULL;
  
  n = 0;
  
  while ((*p>='a'&&*p<='z')||(*p>='A'&&*p<='Z')) {
    buff[n] = *p;
    p++; n++;
  }

  if (n==0) return NULL;

  *next = p;
  buff[n]=0;

  return strdup(buff);
}

int dict_MakeReference(char * paragraph,E2C_Dictionary *DICT,E2C_Dictionary *REF)
{
  int id;
  char * voc;
  char * next = paragraph;
  
  do {
    voc = ref_findword(next,&next);
    if (voc!=NULL) {
      id = dict_SearchEnglish(DICT,voc);
      if (id>=0) ref_add(REF,DICT->DICT[id]);
      free(voc);
    }
  } while (voc!=NULL);
  return REF->size;
}
